(* Definition of the CPS language *)

Require Import Coq.Arith.Peano_dec.
Require Import Coq.Lists.List.
Require Import "Variable".
Require Import ListFacts.


Section Terms.

(* term *)
Inductive cterm : Set :=
  | ct_sbvar (n : nat)  (* bound variable from the source language *)
  | ct_sfvar (x : var)  (* free variable from the source language *)
  | ct_cbvar (n : nat)  (* bound variable introduced in CPS conversion *)
  | ct_cfvar (x : var)  (* free variable introduced in CPS conversion *)
  | ct_unit
  | ct_sabs (t : cterm)
  | ct_cabs (t : cterm)
  | ct_app (tf ta : cterm).

(* substitute bound variable n in term t with term t' *)
Fixpoint open_cterm_s_n (n : nat) (t' t : cterm) : cterm :=
  match t with
  | ct_sbvar n1 =>
      if eq_nat_dec n n1 then t' else ct_sbvar n1
  | ct_sfvar x =>
      ct_sfvar x
  | ct_cbvar n1 =>
      ct_cbvar n1
  | ct_cfvar x =>
      ct_cfvar x
  | ct_unit =>
      ct_unit
  | ct_sabs t1 =>
      ct_sabs (open_cterm_s_n (S n) t' t1)
  | ct_cabs t1 =>
      ct_cabs (open_cterm_s_n n t' t1)
  | ct_app t1 t2 =>
      ct_app (open_cterm_s_n n t' t1) (open_cterm_s_n n t' t2)
  end.
Definition open_cterm_s : cterm -> cterm -> cterm :=
  open_cterm_s_n O.

(* substitute bound variable n in term t with term t' *)
Fixpoint open_cterm_c_n (n : nat) (t' t : cterm) : cterm :=
  match t with
  | ct_sbvar n1 =>
      ct_sbvar n1
  | ct_sfvar x =>
      ct_sfvar x
  | ct_cbvar n1 =>
      if eq_nat_dec n n1 then t' else ct_cbvar n1
  | ct_cfvar x =>
      ct_cfvar x
  | ct_unit =>
      ct_unit
  | ct_sabs t1 =>
      ct_sabs (open_cterm_c_n n t' t1)
  | ct_cabs t1 =>
      ct_cabs (open_cterm_c_n (S n) t' t1)
  | ct_app t1 t2 =>
      ct_app (open_cterm_c_n n t' t1) (open_cterm_c_n n t' t2)
  end.
Definition open_cterm_c : cterm -> cterm -> cterm :=
  open_cterm_c_n O.

End Terms.


Section Types.

(* type *)
Inductive ctype : Set :=
  | cy_unit
  | cy_func (TA TR : ctype)
  | cy_bottom.

End Types.


Section Contexts.

(* typing context *)
Definition cctxt : Set :=
  list (var * ctype).

Definition domain_of_cctxt : cctxt -> list var :=
  map (@fst _ _).

End Contexts.


Section Rules.

Inductive cterm_typing : cctxt -> cctxt -> cterm -> ctype -> Prop :=
  | ctt_sfvar :
      forall (Cs Cc : cctxt) (x : var) (T : ctype),
      in' x T Cs ->
      cterm_typing Cs Cc (ct_sfvar x) T
  | ctt_cfvar :
      forall (Cs Cc : cctxt) (x : var) (T : ctype),
      in' x T Cc ->
      cterm_typing Cs Cc (ct_cfvar x) T
  | ctt_unit :
      forall Cs Cc : cctxt,
      cterm_typing Cs Cc ct_unit cy_unit
  | ctt_sabs :
      forall (Cs Cc : cctxt) (l : list var) (TA TR : ctype) (t : cterm),
      (forall x : var, ~ In x l ->
       cterm_typing ((x, TA) :: Cs) Cc (open_cterm_s (ct_sfvar x) t) TR) ->
      cterm_typing Cs Cc (ct_sabs t) (cy_func TA TR)
  | ctt_cabs :
      forall (Cs Cc : cctxt) (l : list var) (TA TR : ctype) (t : cterm),
      (forall x : var, ~ In x l ->
       cterm_typing Cs ((x, TA) :: Cc) (open_cterm_c (ct_cfvar x) t) TR) ->
      cterm_typing Cs Cc (ct_cabs t) (cy_func TA TR)
  | ctt_app :
      forall (Cs Cc : cctxt) (TA TR : ctype) (tf ta : cterm),
      cterm_typing Cs Cc tf (cy_func TA TR) ->
      cterm_typing Cs Cc ta TA ->
      cterm_typing Cs Cc (ct_app tf ta) TR.

End Rules.


Bind Scope ctype_scope with ctype.
Delimit Scope ctype_scope with ctype.
Notation "! T" := (cy_func T cy_bottom)
  (at level 35) : ctype_scope.
